// chronock.cpp - assert monotonic time
// v1.2 // 2018-06-29 // ed <s@ocv.me> // MIT licensed

// g++ -O9 --std=c++11 -Weffc++ -Wall -Wextra -Werror -Wfatal-errors chronock.cpp && ./a.out
// pid=$(pidof a.out); while true; do for x in 0 2 4 6 4 2 0 6 0 4; do taskset -pac $x $pid; done; done 
// cat /sys/devices/system/clocksource/clocksource0/{current_clocksource,available_clocksource}

//    a/mono  b/asmb  c/raw   d/gtod  e/real  f/stdy  g/hrc
// 1  0.0926  0.4349  1.3566  0.1348  0.0936  1.3338  1.3153  a e d b g f c
// 2  0.0657  0.3467  1.2839  0.1057  0.0662  1.3048  1.3203  a e d b c f g
// 3  0.1605  0.5793  0.7628  0.1871  0.1603  0.7702  0.8055  e a d b c f g

//    cpu        g++    linux            distro      clock
// 1: i5-6300HQ, 6.4.0, 4.14.52 vanilla, alpine 3.7, [tsc] hpet acpi_pm
// 2: E3-1230v3, 6.4.0, 4.9.71 grsec,    alpine 3.7, [tsc] hpet acpi_pm
// 3: E5506,     4.8.5, 3.10.0-693,      centos 7.4, [tsc] hpet acpi_pm


#include <iostream>
#include <iomanip>
#include <vector>
#include <map>
#include <chrono>      // chrono::steady_clock
#include <time.h>      // clock_gettime
#include <sys/time.h>  // gettimeofday
#include <cstring>     // memset

#define IN_ORDER_EXEC \
	asm volatile("CPUID":"=a"(eax),"=b"(ebx),"=c"(ecx),"=d"(edx):"0"(lv):"memory");

#define NANO 1000000000ULL


bool test_a(uint64_t& sum)
{
	sum = 0;
	struct timespec tp;
	for (int a = 0; a < 1024 * 1024 * 2; a++)
	{
		clock_gettime(CLOCK_MONOTONIC, &tp);  uint64_t t1 = (uint64_t)tp.tv_sec * NANO + (uint64_t)tp.tv_nsec;
		clock_gettime(CLOCK_MONOTONIC, &tp);  uint64_t t2 = (uint64_t)tp.tv_sec * NANO + (uint64_t)tp.tv_nsec;
		sum += t2 - t1;
		if (t2 < t1)
		{
			sum = t1 - t2;
			return false;
		}
	}
	return true;
}


bool test_b(uint64_t& sum)
{
	sum = 0;
	int eax, ebx, ecx, edx, lv = 0;
	for (int a = 0; a < 1024 * 1024 * 2; a++)
	{
		IN_ORDER_EXEC;
		struct timespec tp1;
		memset(&tp1, 0, sizeof(tp1));
		IN_ORDER_EXEC;
		clock_gettime(CLOCK_MONOTONIC, &tp1);
		uint64_t t1 =
			(uint64_t)tp1.tv_sec * NANO +
			(uint64_t)tp1.tv_nsec;
		
		IN_ORDER_EXEC;
		struct timespec tp2;
		memset(&tp2, 0, sizeof(tp2));
		IN_ORDER_EXEC;
		clock_gettime(CLOCK_MONOTONIC, &tp2);
		uint64_t t2 =
			(uint64_t)tp2.tv_sec * NANO +
			(uint64_t)tp2.tv_nsec;
		
		IN_ORDER_EXEC;
		sum += t2 - t1;
		if (t2 < t1)
		{
			sum = t1 - t2;
			return false;
		}
	}
	return true;
}


bool test_c(uint64_t& sum)
{
	sum = 0;
	struct timespec tp;
	for (int a = 0; a < 1024 * 1024 * 2; a++)
	{
		memset(&tp, 0, sizeof(tp));  clock_gettime(CLOCK_MONOTONIC_RAW, &tp);  uint64_t t1 = (uint64_t)tp.tv_sec * NANO + (uint64_t)tp.tv_nsec;
		clock_gettime(CLOCK_MONOTONIC_RAW, &tp);  uint64_t t2 = (uint64_t)tp.tv_sec * NANO + (uint64_t)tp.tv_nsec;
		sum += t2 - t1;
		if (t2 < t1)
		{
			sum = t1 - t2;
			return false;
		}
	}
	return true;
}


bool test_d(uint64_t& sum)
{
	sum = 0;
	struct timeval tv;
	for (int a = 0; a < 1024 * 1024 * 2; a++)
	{
		gettimeofday(&tv, NULL);  uint64_t t1 = (uint64_t)tv.tv_sec * NANO + (uint64_t)tv.tv_usec * 1000ULL;
		gettimeofday(&tv, NULL);  uint64_t t2 = (uint64_t)tv.tv_sec * NANO + (uint64_t)tv.tv_usec * 1000ULL;
		sum += t2 - t1;
		if (t2 < t1)
		{
			sum = t1 - t2;
			return false;
		}
	}
	return true;
}


bool test_e(uint64_t& sum)
{
	sum = 0;
	struct timespec tp;
	for (int a = 0; a < 1024 * 1024 * 2; a++)
	{
		clock_gettime(CLOCK_REALTIME, &tp);  uint64_t t1 = (uint64_t)tp.tv_sec * NANO + (uint64_t)tp.tv_nsec;
		clock_gettime(CLOCK_REALTIME, &tp);  uint64_t t2 = (uint64_t)tp.tv_sec * NANO + (uint64_t)tp.tv_nsec;
		sum += t2 - t1;
		if (t2 < t1)
		{
			sum = t1 - t2;
			return false;
		}
	}
	return true;
}


bool test_f(uint64_t& sum)
{
	sum = 0;
	for (int a = 0; a < 1024 * 1024 * 2; a++)
	{
		auto t1 = std::chrono::steady_clock::now();
		auto t2 = std::chrono::steady_clock::now();
		sum += std::chrono::duration_cast<std::chrono::nanoseconds>(t2-t1).count();
		if (t2 < t1)
		{
			sum = std::chrono::duration_cast<std::chrono::nanoseconds>(t1-t2).count();
			return false;
		}
	}
	return true;
}


bool test_g(uint64_t& sum)
{
	double dsum = 0;
	for (int a = 0; a < 1024 * 1024 * 2; a++)
	{
		std::chrono::high_resolution_clock::time_point t1 = std::chrono::high_resolution_clock::now();
		std::chrono::high_resolution_clock::time_point t2 = std::chrono::high_resolution_clock::now();
		dsum += std::chrono::duration_cast<std::chrono::duration<double>>(t2-t1).count();
		if (t2 < t1)
		{
			sum = std::chrono::duration_cast<std::chrono::nanoseconds>(t1-t2).count();
			return false;
		}
	}
	sum = dsum * (double)NANO;
	return true;
}


bool test_h(uint64_t& sum)
{
	double dsum = 0;
	for (int a = 0; a < 1024 * 1024 * 2; a++)
	{
		auto t1 = std::chrono::steady_clock::now();
		auto t2 = std::chrono::steady_clock::now();
		dsum += std::chrono::duration_cast<std::chrono::duration<double>>(t2-t1).count();
		if (t2 < t1)
		{
			sum = std::chrono::duration_cast<std::chrono::nanoseconds>(t1-t2).count();
			return false;
		}
	}
	sum = dsum * (double)NANO;
	return true;
}


bool test_i(uint64_t& sum)
{
	double dsum = 0;
	for (int a = 0; a < 1024 * 1024 * 2; a++)
	{
		auto t1 = std::chrono::high_resolution_clock::now();
		auto t2 = std::chrono::high_resolution_clock::now();
		dsum += std::chrono::duration_cast<std::chrono::duration<double>>(t2-t1).count();
		if (t2 < t1)
		{
			sum = std::chrono::duration_cast<std::chrono::nanoseconds>(t1-t2).count();
			return false;
		}
	}
	sum = dsum * (double)NANO;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

int main(int, char**)
{
	typedef bool (*Test)(uint64_t&);
	std::vector<std::pair<std::string, Test>> tests;
	tests.emplace_back("a", &test_a);
	tests.emplace_back("b", &test_b);
	tests.emplace_back("c", &test_c);
	tests.emplace_back("d", &test_d);
	tests.emplace_back("e", &test_e);
	tests.emplace_back("f", &test_f);
	tests.emplace_back("g", &test_g);
	//tests.emplace_back("h", &test_h);
	//tests.emplace_back("i", &test_h);
	
	uint32_t iterations = 0;
	std::map<std::string, double> spent;
	for (const auto& pair : tests)
		spent[pair.first] = 0;
	
	std::cout << std::fixed << std::setprecision(4);
	
	while (true)
	{
		iterations++;
		std::cout << "\033[32m\n";
		for (const auto& pair : tests)
			std::cout << pair.first << "  ";
		
		std::cout << "\033[0m\n";
		for (auto it = tests.begin(); it != tests.end(); it++)
		{
			uint64_t sum;
			std::string name = (*it).first;
			Test func = (*it).second;
			
			auto t1 = std::chrono::steady_clock::now();
			if (!(func)(sum))
			{
				std::cout << "\033[1;31mtest \033[37m" << name << "\033[31m is not monotonous: \033[0mdecreased by " << sum << " ns\n";
				tests.erase(it);
				break;
			}
			auto t2 = std::chrono::steady_clock::now();
			double td = std::chrono::duration_cast<std::chrono::duration<double>>(t2 - t1).count();
			spent[name] += td;
			
			std::cout << "\033[1;33m" << name << ": \033[0m" << td
				<< "  \033[13G" << spent[name]/iterations
				<< "  \033[22G" << sum << std::endl;
		}
		std::cout << "\033[1;30mT: this_it  tot_avg  returned sum\n";
		
		for (const auto& pair : tests)
			std::cout << spent[pair.first]/iterations << "  ";
		
		std::cout << "\033[0m\n";
	}
	return 0;
}

